// RigidBodyPlugin.cpp by d.j.peters (Joshy)
/*
#ifdef  _MSC_VER
  extern "C" long _ftol( double ); //defined by VC6 C libs
  extern "C" long _ftol2( double dblSource ) { return _ftol( dblSource ); }
#endif
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL/glu.h>

#include <Basic4GLDLLInterface.h>
#include <Basic4GLStandardObjects.h>
#include "Basic4GLPluginMacros.h"
#include "rigidbody2.h"



///////////////////////////////////////////////////////////////////////////////
// RigidBodyPlugin
//


// Global variables.
char                  * pluginError     = NULL;
IB4GLCompiler         * gpCompiler      = NULL;
float m[16];
float t1[3],t2[3],t3[3],t9[9];
float o1[3],o2[3],o3[3],o9[9];

RigidBody             * gpRB=NULL;

// registered types
int ghRB   = 0; // RigidBody

// get set return classes

// RigidBody
#define GRB(n)  GT(ghRB,n,&gpRB)
#define SRB(n)  ST(ghRB,n,&gpRB)
#define RRB()   STR(ghRB,&gpRB)
#define RPRB(p) STR(ghRB,&p)


class RigidBodyPlugin : public IDLL_Basic4GL_Plugin {
    public:
    RigidBodyPlugin() { ; }
    virtual bool DLLFUNC Load(IDLL_Basic4GL_FunctionRegistry &registry, bool isStandaloneExe);
    virtual bool DLLFUNC Start();
    virtual void DLLFUNC End();
    virtual void DLLFUNC Unload();
    virtual void DLLFUNC GetError(char *error);
    virtual void DLLFUNC Pause();
    virtual void DLLFUNC Resume();
    virtual void DLLFUNC DelayedResume();
    virtual void DLLFUNC ProcessMessages();
};
///////////////////////////////////////////////////////////////////////////////
// DLL exported functions
extern "C" {
  /// Query function
  DLLExport int DLLFUNC Basic4GL_Query(char *details, int *major, int *minor) {
    char *detailSrc = "for Basic4GL";
    for (int i = 0; detailSrc[i] != 0; i++){
      details[i] = detailSrc[i];
    }
    *major = 1; *minor = 25;
    return BASIC4GL_DLL_VERSION;
  }

  DLLExport IDLL_Basic4GL_Plugin *DLLFUNC Basic4GL_Init() {
    return new RigidBodyPlugin();
  }
}


///////////////////////////////////////////////////////////////////////////////
// Runtime functions
// 

// #############
// # RigidBody #
// #############
DPF(RigidBodySetStepSize){ // (stepsize#)
  RigidBody::setStepSize(GF(1));
}
DPF(RigidBodyGetStepSize){ // stepsize#=()
  SFR(RigidBody::RKh);
}

DPF(RigidBodySetGravity){ // (gravity#)
  RigidBody::setGravity(GF(1));
}
DPF(RigidBodyGetGravity){ // gravity#=()
  SFR(RigidBody::g);
}
DPF(CreateRigidBodyBox){ // obj=(rx#,ry#,rz#,mass#)
  gpRB = new RigidBody; gpRB->setBox(GF(4),GF(2),GF(3), GF(1)); RRB();
}
DPF(CreateRigidBodyCube){ // obj=(rx#,mass#)
  gpRB = new RigidBody; gpRB->setBox(GF(2),GF(2),GF(2), GF(1)); RRB();
}
DPF(CreateRigidBodySphere){ // obj=(radius#, mass#)
  gpRB = new RigidBody; gpRB->setSphere(GF(2),GF(1)); RRB();
}
DPF(CreateRigidBodyCylinder){ // obj=(radius#,height#,mass#)
  gpRB = new RigidBody; gpRB->setCylinder(GF(3),GF(2),GF(1)); RRB();
}


DPF(DestroyRigidBody) {// (obj)
  GRB(1);if (gpRB==0) return;
  delete gpRB;gpRB=0;SRB(1);
}

DPF(RigidBodyCombine){ // (obj,objOther,rotPos#(),linPos#())
  RigidBody * pOther=NULL;
  GRB(4);            if (gpRB==0)   return;
  GT(ghRB,3,&pOther);if (pOther==0) return;
  Vector3 rotPos;GFA(2,&rotPos.x,3);
  Vector3 linPos;GFA(1,&linPos.x,3);
  Matrix3x3  mRotPos; mRotPos.rpyRotation(rotPos.x,rotPos.z,rotPos.y); // Z up
  gpRB->combine(*pOther,mRotPos,linPos);
}
DPF(RigidBodyDiagonalise){ // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->diagonalise();
}

DPF(RigidBodyInitialise){ // (obj, linPos#(),linVel#(),rotPos#(),rotVel#().linDamping#,rotDamping#
  GRB(7);if (gpRB==0) return;
  //                           x  ,  z  ,  y
  GFA(6,t1,3);Vector3 linPos(t1[0],t1[2],t1[1]);
  GFA(5,t1,3);Vector3 linVel(t1[0],t1[2],t1[1]);
  GFA(4,t1,3);Vector3 rotPos(t1[0],t1[2],t1[1]);
  GFA(3,t1,3);Vector3 rotVel(t1[0],t1[2],t1[1]);
  Matrix3x3  mRotPos; mRotPos.rpyRotation(rotPos.x,rotPos.y,rotPos.y);
  Quaternion qRotPos; qRotPos.fromMatrix(mRotPos);
  gpRB->initialise(linPos,linVel,qRotPos,rotVel,GF(2),GF(1));
}

// simualtion
DPF(RigidBodyRKInit) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->RKInit();
}

DPF(RigidBodyEulerStep) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->eulerStep();
}
DPF(RigidBodyRKStep) { // (obj,step%)
  GRB(2);if (gpRB==0) return;
  int step=GI(1);
  switch (step) {
  case 1: gpRB->RKStep1();break;
  case 2: gpRB->RKStep2();break;
  case 3: gpRB->RKStep3();break;
  case 4: gpRB->RKStep4();break;
  }
  
}

DPF(RigidBodyRKStep1) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->RKStep1();
}

DPF(RigidBodyRKStep2) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->RKStep2();
}

DPF(RigidBodyRKStep3) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->RKStep3();
}

DPF(RigidBodyRKStep4) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->RKStep4();
}
DPF(RigidBodyUndo) { // (obj)
  GRB(1);if (gpRB==0) return;
  gpRB->undo();
}


// Force
DPF(RigidBodyAddWorldWorldForce) { // (obj, forceXDir#,forceYDir#,forceZDir#, px#,py#,pz#)
  GRB(7);if (gpRB==0) return;
  gpRB->addWorldWorldForce( Vector3(GF(6),GF(4),GF(5)),Vector3(GF(3),GF(1),GF(2)) ); // Z is up
}
DPF(RigidBodyAddWorldWorldForceA) {// (obj, forceDir#(), position#(), )
  GRB(3);if (gpRB==0) return;
  GFA(2,t1,3);GFA(1,t2,3);
  gpRB->addWorldWorldForce( Vector3(t1[0],t1[2],t1[1]),Vector3(t2[0],t2[2],t2[1]) ); // Z is up
}


DPF(RigidBodyAddWorldBodyForce) { // (obj, forceXDir#,forceYDir#,forceZDir#, px#,py#,pz#)
  GRB(7);if (gpRB==0) return;
  gpRB->addWorldBodyForce( Vector3(GF(6),GF(4),GF(5)), Vector3(GF(3),GF(1),GF(2)) ); // Z is up
}
DPF(RigidBodyAddWorldBodyForceA) {// (obj, forceDir#(), position#(), )
  GRB(3);if (gpRB==0) return;
  GFA(2,t1,3);GFA(1,t2,3);
  gpRB->addWorldBodyForce( Vector3(t1[0],t1[2],t1[1]),Vector3(t2[0],t2[2],t2[1]) ); // Z is up
}

DPF(RigidBodyAddBodyBodyForce) { // (obj, forceXDir#,forceYDir#,forceZDir#, px#,py#,pz#)
  GRB(7);if (gpRB==0) return;
  gpRB->addBodyBodyForce( Vector3(GF(6),GF(4),GF(5)), Vector3(GF(3),GF(1),GF(2)) ); // Z is up
}
DPF(RigidBodyAddBodyBodyForceA) {// (obj, forceDir#(), position#(), )
  GRB(3);if (gpRB==0) return;
  GFA(2,t1,3);GFA(1,t2,3);
  gpRB->addBodyBodyForce( Vector3(t1[0],t1[2],t1[1]),Vector3(t2[0],t2[2],t2[1]) ); // Z is up
}

// Torque
DPF(RigidBodyAddWorldTorque){ // (obj, xTorque#,yTorque#,zTorque#)
  GRB(4);if (gpRB==0) return;
  gpRB->addWorldTorque(Vector3(GF(3),GF(1),GF(2))); // Z up
}
DPF(RigidBodyAddWorldTorqueA){// (obj, torque#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->addWorldTorque(Vector3(t1[0],t1[2],t1[1])); // Z is up
}

DPF(RigidBodyAddBodyTorque){ // (obj, xTorque#,yTorque#,zTorque#)
  GRB(4);if (gpRB==0) return;
  gpRB->addWorldTorque(Vector3(GF(3),GF(1),GF(2))); // Z up
}
DPF(RigidBodyAddBodyTorqueA){// (obj, torque#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->addWorldTorque(Vector3(t1[0],t1[2],t1[1])); // Z is up
}


// properties
DPF(RigidBodyGetMass){ // mass#=(obj)
  GRB(1);if (gpRB==0) return;
  SFR(gpRB->mass);
}

DPF(RigidBodyGetEnergy){
  GRB(1);if (gpRB==0) return;
  SFR(gpRB->energy());
}

DPF(RigidBodySetUserData){
  GRB(2);if (gpRB==0) return;
  gpRB->setUserData(GI(1));
}
DPF(RigidBodyGetUserData){
  GRB(1);if (gpRB==0) return;
  SIR(gpRB->getUserData());
}

DPF(RigidBodySetLinPos){ // (obj,px#,py#,pz#)
  GRB(4);if (gpRB==0) return;
  gpRB->linPos.x=GF(3);
  gpRB->linPos.z=GF(2); // Z is up
  gpRB->linPos.y=GF(1);
}
DPF(RigidBodySetLinPosA){ // (obj,pos#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->linPos.x=t1[0];
  gpRB->linPos.y=t1[2]; // Z is up
  gpRB->linPos.z=t1[1];
}
DPF(RigidBodyGetLinPos){ // pos#()=(obj)
  GRB(1);if (gpRB==0) return;
  t1[0]=gpRB->linPos.x;
  t1[1]=gpRB->linPos.z; // Z is up
  t1[2]=gpRB->linPos.y;
  SFAR(t1,3);
}


DPF(RigidBodySetLinVel){ // (obj,vx#,vy#,vz#)
  GRB(4);if (gpRB==0) return;
  gpRB->linVel.x=GF(3);
  gpRB->linVel.y=GF(1); // Z is up
  gpRB->linVel.z=GF(2);
}
DPF(RigidBodySetLinVelA){ // (obj,linVel#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->linVel.x=t1[0];
  gpRB->linVel.y=t1[2]; // Z is up
  gpRB->linVel.z=t1[1];
}
DPF(RigidBodyGetLinVel){ // linVel#()=(obj)
  GRB(1);if (gpRB==0) return;
  t1[0]=gpRB->linVel.x;
  t1[1]=gpRB->linVel.z; // Z is up
  t1[2]=gpRB->linVel.y;
  SFAR(t1,3);
}


DPF(RigidBodySetRotVel){ // (obj,vx#,vy#,vz#)
  GRB(4);if (gpRB==0) return;
  gpRB->rotVel.x=GF(3);
  gpRB->rotVel.y=GF(1); // Z is up
  gpRB->rotVel.z=GF(2);
}
DPF(RigidBodySetRotVelA){ // (obj,velocity#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->rotVel.x=t1[0];
  gpRB->rotVel.y=t1[2]; // Z is up
  gpRB->rotVel.z=t1[1];
}
DPF(RigidBodyGetRotVel){ // rotVel#()=(obj)
  GRB(1);if (gpRB==0) return;
  t1[0]=gpRB->rotVel.x;
  t1[1]=gpRB->rotVel.z; // Z is up
  t1[2]=gpRB->rotVel.y;
  SFAR(t1,3);
}

DPF(RigidBodySetLinForce){ // (obj,px#,py#,pz#)
  GRB(4);if (gpRB==0) return;
  gpRB->linForce.x=GF(3);
  gpRB->linForce.y=GF(1); // Z is up
  gpRB->linForce.z=GF(2);
}
DPF(RigidBodySetLinForceA){ // (obj,force#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->linForce.x=t1[0];
  gpRB->linForce.y=t1[2]; // Z is up
  gpRB->linForce.z=t1[1];
}
DPF(RigidBodyGetLinForce){ // linForce#()=(obj)
  GRB(1);if (gpRB==0) return;
  t1[0]=gpRB->linForce.x;
  t1[1]=gpRB->linForce.z; // Z is up
  t1[2]=gpRB->linForce.y;
  SFAR(t1,3);
}

DPF(RigidBodySetRotForce){ // (obj,px#,py#,pz#)
  GRB(4);if (gpRB==0) return;
  gpRB->rotForce.x=GF(3);
  gpRB->rotForce.y=GF(1); // Z is up
  gpRB->rotForce.z=GF(2);
}
DPF(RigidBodySetRotForceA){ // (obj,force#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  gpRB->rotForce.x=t1[0];
  gpRB->rotForce.y=t1[2]; // Z is up
  gpRB->rotForce.z=t1[1];
}
DPF(RigidBodyGetRotForce){ // linForce#()=(obj)
  GRB(1);if (gpRB==0) return;
  t1[0]=gpRB->rotForce.x;
  t1[1]=gpRB->rotForce.z; // Z is up
  t1[2]=gpRB->rotForce.y;
  SFAR(t1,3);
}




// Get position of body. Converts diagonalised coords back to gfx coords :
DPF(RigidBodyGetWorldMatrix){ // m#()() = (obj)
  GRB(1);if (gpRB==0) return;
  Vector3   linPos;
  Matrix3x3 rotPos;
  rotPos.mult(gpRB->rotPos, gpRB->diagRotPos);
  linPos.transform(gpRB->diagLinPos, gpRB->rotPos, gpRB->linPos);
  m[ 0]=rotPos.m[0]; m[ 1]=rotPos.m[1]; m[ 2]=rotPos.m[2]; m[ 3]=0;
  m[ 4]=rotPos.m[3]; m[ 5]=rotPos.m[4]; m[ 6]=rotPos.m[5]; m[ 7]=0;
  m[ 8]=rotPos.m[6]; m[ 9]=rotPos.m[7]; m[10]=rotPos.m[8]; m[11]=0;
  m[12]=linPos.m[0]; m[13]=linPos.m[2]; m[14]=linPos.m[1]; m[15]=1; // !!! z=up
  SFAR2(m,4,4);
}

DPF(RigidBodyMultOpenGLWorldMatrix){ // (obj)
  GRB(1);if (gpRB==0) return;
  Vector3   linPos;
  Matrix3x3 rotPos;
  rotPos.mult(gpRB->rotPos, gpRB->diagRotPos);
  linPos.transform(gpRB->diagLinPos, gpRB->rotPos, gpRB->linPos);
  m[ 0]=rotPos.m[0]; m[ 1]=rotPos.m[1]; m[ 2]=rotPos.m[2]; m[ 3]=0;
  m[ 4]=rotPos.m[3]; m[ 5]=rotPos.m[4]; m[ 6]=rotPos.m[5]; m[ 7]=0;
  m[ 8]=rotPos.m[6]; m[ 9]=rotPos.m[7]; m[10]=rotPos.m[8]; m[11]=0;
  m[12]=linPos.m[0]; m[13]=linPos.m[2]; m[14]=linPos.m[1]; m[15]=1; // !!! z=up
  glMultMatrixf(m);

}


DPF(RigidBodyGetMatrix){ // m#()() = (obj)
  GRB(1);if (gpRB==0) return;
  m[ 0]=gpRB->rotPos.m[0]; m[ 1]=gpRB->rotPos.m[1]; m[ 2]=gpRB->rotPos.m[2]; m[ 3]=0;
  m[ 4]=gpRB->rotPos.m[3]; m[ 5]=gpRB->rotPos.m[4]; m[ 6]=gpRB->rotPos.m[5]; m[ 7]=0;
  m[ 8]=gpRB->rotPos.m[6]; m[ 9]=gpRB->rotPos.m[7]; m[10]=gpRB->rotPos.m[8]; m[11]=0;
  m[12]=gpRB->linPos.m[0]; m[13]=gpRB->linPos.m[2]; m[14]=gpRB->linPos.m[1]; m[15]=1; // !!! z=up
  SFAR2(m,4,4);
}

DPF(RigidBodyMultOpenGLMatrix){ // (obj)
  GRB(1);if (gpRB==0) return;
  m[ 0]=gpRB->rotPos.m[0]; m[ 1]=gpRB->rotPos.m[1]; m[ 2]=gpRB->rotPos.m[2]; m[ 3]=0;
  m[ 4]=gpRB->rotPos.m[3]; m[ 5]=gpRB->rotPos.m[4]; m[ 6]=gpRB->rotPos.m[5]; m[ 7]=0;
  m[ 8]=gpRB->rotPos.m[6]; m[ 9]=gpRB->rotPos.m[7]; m[10]=gpRB->rotPos.m[8]; m[11]=0;
  m[12]=gpRB->linPos.m[0]; m[13]=gpRB->linPos.m[2]; m[14]=gpRB->linPos.m[1]; m[15]=1; // !!! z=up

  glMultMatrixf(m);

}

DPF(RigidBodyTransformVec3) { // vec3#()=(obj,vec3#())
  GRB(2);if (gpRB==0) return;
  GFA(1,t1,3);
  Vector3 oldVec3(t1[0],t1[2],t1[1]);
  Vector3 newVec3;
  newVec3.transform(oldVec3,gpRB->diagRotPos,gpRB->diagLinPos);
  t2[0]=newVec3.x;
  t2[2]=newVec3.y; // Z up
  t2[1]=newVec3.z;
  SFAR(t2,3);
}






///////////////////////////////////////////////////////////////////////////////
//	TokamakPlugin

bool DLLFUNC RigidBodyPlugin::Load(IDLL_Basic4GL_FunctionRegistry &registry, bool isStandaloneExe) {
  // Initialise DLL stuff
  // Register shared interfaces
  // Register constants
  // Register user defined data types (UDT's)
  // Register runtime functions
  isStandaloneExe=isStandaloneExe;
  // Fetch  interface
  
  /*
    gpCompiler=(IB4GLCompiler*)registry.FetchInterface("IB4GLCompiler",
                                                     IB4GLCOMPILER_MAJOR,
                                                     IB4GLCOMPILER_MINOR);
  */

  RigidBody::setStepSize(1.0f/100.0f);
  RigidBody::setGravity(9.81f);


  // register const's

  // register classes
   ghRB   = RT(Rigidbody,     0,1);RTIF(THIS);
 
  // #############
  // # RigidBody #
  // #############
  RVF (RigidBodySetStepSize);AFP; // (stepsize#)
  RFF (RigidBodyGetStepSize);     // stepsize#=()
  
  RVF (RigidBodySetGravity); AFP; // (gravity#)
  RFF (RigidBodyGetGravity);      // gravity#=()

  RTF (CreateRigidBodyBox,ghRB);            AFP;AFP;AFP; AFP; // obj=(halfXSize#,halfYSize#,halfZSize#,mass#)
  RTF (CreateRigidBodyCube,ghRB);           AFP;AFP;          // obj=(radius#,mass#)
  RTF (CreateRigidBodySphere,ghRB);         AFP;         AFP; // obj=(radius#,mass#)
  RTF (CreateRigidBodyCylinder,ghRB);       AFP;AFP;     AFP; // obj=(radius#,height#,mass#)

  RVF (DestroyRigidBody);ATP(ghRB);                           // (obj)
                     
  RVF (RigidBodyCombine);ATP(ghRB);ATP(ghRB);AFAP(1);  AFAP(1); // (obj,objOther,rotPos#(),inPos#())
  RVF (RigidBodyDiagonalise);ATP(ghRB);                       // (obj) 
                            // (obj,  linPos#(),linVel#(),rotPos#(),rotVel#().linDamping#,rotDamping#
  RVF (RigidBodyInitialise);ATP(ghRB);AFAP(1);  AFAP(1);  AFAP(1);  AFAP(1);  AFP;        AFP; 

  // simulation
  RVF (RigidBodyRKInit);    ATP(ghRB); // (obj)
  RVF (RigidBodyEulerStep); ATP(ghRB); // (obj)
  RVF (RigidBodyRKStep);    ATP(ghRB); // (obj,step%)

  RVF (RigidBodyRKStep1);   ATP(ghRB); // (obj)
  RVF (RigidBodyRKStep2);   ATP(ghRB); // (obj)
  RVF (RigidBodyRKStep3);   ATP(ghRB); // (obj)
  RVF (RigidBodyRKStep4);   ATP(ghRB); // (obj)
  
  RVF (RigidBodyUndo);      ATP(ghRB); // (obj)

  // Force
  RVF (RigidBodyAddWorldWorldForce);ATP(ghRB);AFP;AFP;AFP; AFP;AFP;AFP; // (obj, fx#,fy#,fz#, px#,py#,pz#)
  RVF2(RigidBodyAddWorldWorldForce,RigidBodyAddWorldWorldForceA);ATP(ghRB);AFAP(1),AFAP(1); // (obj, force#(), forceDir#())

  RVF (RigidBodyAddWorldBodyForce);ATP(ghRB);AFP;AFP;AFP; AFP;AFP;AFP; // (obj, fx#,fy#,fz#, px#,py#,pz#)
  RVF2(RigidBodyAddWorldBodyForce,RigidBodyAddWorldBodyForceA);ATP(ghRB);AFAP(1);AFAP(1); // (obj, force#(), forceDir#())

  RVF (RigidBodyAddBodyBodyForce);ATP(ghRB);AFP;AFP;AFP; AFP;AFP;AFP; // (obj, fx#,fy#,fz#, px#,py#,pz#)
  RVF2(RigidBodyAddBodyBodyForce,RigidBodyAddBodyBodyForceA);ATP(ghRB);AFAP(1);AFAP(1); // (obj, force#(), forceDir#())

  // Torque
  RVF (RigidBodyAddWorldTorque);ATP(ghRB);AFP;AFP;AFP;  // (obj, xTorque#,yTorque#,zTorque#)
  RVF2(RigidBodyAddWorldTorque,RigidBodyAddWorldTorqueA);ATP(ghRB);AFAP(1); // (obj, torque#())

  RVF (RigidBodyAddBodyTorque);ATP(ghRB);AFP;AFP;AFP;  // (obj, xTorque#,yTorque#,zTorque#)
  RVF2(RigidBodyAddBodyTorque,RigidBodyAddBodyTorqueA);ATP(ghRB);AFAP(1); // (obj, torque#())

  // Properties
  RFF (RigidBodyGetMass);     ATP(ghRB); // mass#=(obj)

  RVF (RigidBodySetUserData); ATP(ghRB); // (obj,ud%)
  RIF (RigidBodyGetUserData); ATP(ghRB); // ud%=(obj)

  RVF (RigidBodySetLinPos);                    ATP(ghRB);AFP;AFP;AFP; // (obj,linXPos#,linYPos#,linZPos#)
  RVF2(RigidBodySetLinPos,RigidBodySetLinPosA);ATP(ghRB);AFAP(1);     // (obj,linPos#())
  RFAF(RigidBodyGetLinPos,1);                  ATP(ghRB);             // linPos#()=(obj)
  
  RVF (RigidBodySetLinVel);                    ATP(ghRB);AFP;AFP;AFP; // (obj,linXVel#,linYVel#,linZVel#)
  RVF2(RigidBodySetLinVel,RigidBodySetLinVelA);ATP(ghRB);AFAP(1);     // (obj,linVel#())
  RFAF(RigidBodyGetLinVel,1);                  ATP(ghRB);             // linVel#()=(obj)

  RVF (RigidBodySetRotVel);                    ATP(ghRB);AFP;AFP;AFP; // (obj,rotXVel#,rotYVel#,rotZVel#)
  RVF2(RigidBodySetRotVel,RigidBodySetRotVelA);ATP(ghRB);AFAP(1);     // (obj,rotVel#())
  RFAF(RigidBodyGetRotVel,1);                  ATP(ghRB);             // rotVel#()=(obj)

  RVF (RigidBodySetLinForce);                      ATP(ghRB);AFP;AFP;AFP; // (obj,linXForce#,linYForce#,linZForce#)
  RVF2(RigidBodySetLinForce,RigidBodySetLinForceA);ATP(ghRB);AFAP(1);     // (obj,linForce#())
  RFAF(RigidBodyGetLinForce,1);                    ATP(ghRB);             // linForce#()=(obj)

  RVF (RigidBodySetRotForce);                      ATP(ghRB);AFP;AFP;AFP; // (obj,rotXForce#,rotYForce#,rotZForce#)
  RVF2(RigidBodySetRotForce,RigidBodySetRotForceA);ATP(ghRB);AFAP(1);     // (obj,rotForce#())
  RFAF(RigidBodyGetRotForce,1);                    ATP(ghRB);             // rotForce#()=(obj)
  
  RFAF(RigidBodyGetMatrix,2);          ATP(ghRB);                         // m#()()=(obj)
  RVF (RigidBodyMultOpenGLMatrix);     ATP(ghRB);                         // (obj) 

  RFAF(RigidBodyGetWorldMatrix,2);     ATP(ghRB);                         // m#()()=(obj)
  RVF (RigidBodyMultOpenGLWorldMatrix);ATP(ghRB);                         // (obj) 

  RFAF(RigidBodyTransformVec3,1);      ATP(ghRB);AFAP(1);                 // vec3diagonalized#()=(obj,vec3#())
  
  
  RFF (RigidBodyGetEnergy);   ATP(ghRB);                      // energy#=(obj)
  
  return true;
}
void DLLFUNC RigidBodyPlugin::Unload() {
    // DLL is about to unload.
    // Final cleanup code

    // Delete this object. This is safe because Basic4GL will not access it again.
    delete this;
}
bool DLLFUNC RigidBodyPlugin::Start() {
    // Program is about to start.
    // Any pre-start init code would go here.
    return true;
}
void DLLFUNC RigidBodyPlugin::End() {
    // Program is about to end
    // Any cleanup code would go here

}
void DLLFUNC RigidBodyPlugin::Pause() {
    // Called when the program is paused (i.e. during debugging).
    // A full screen window might hide itself here so that the user can see the
    // Basic4GL edit window (and debug their program)
}
void DLLFUNC RigidBodyPlugin::Resume() {
    // Called when the program is resumed (i.e. during debugging).
    // This is also called when the user "steps"
}
void DLLFUNC RigidBodyPlugin::DelayedResume() {
    // Called when the program is resumed for more than a short "step".
    // This is where a fullscreen window would re-show itself (if it did so in ::Resume()
    // we would get flickering when the user steps through the program)
}
void DLLFUNC RigidBodyPlugin::GetError(char *error) {
  if (pluginError != NULL) {
    for (int i = 0; pluginError[i] != 0; i++){
        error[i] = pluginError[i];
    }
  }
}
void DLLFUNC RigidBodyPlugin::ProcessMessages() {
    // Called periodically (in practice, quite frequently).
    // This would be a good place to process windows messages etc.
}

